from .context import FeatureFlagsContext
from .core import FeatureFlag, FeatureFlags
from .doc import make_napoleon_doc, make_sphinx_doc
from .parser import feature_flags_from_environ, parse_feature_flags_string

__version__ = "0.1.0"
__all__ = [
    "FeatureFlag",
    "FeatureFlags",
    "FeatureFlagsContext",
    "make_napoleon_doc",
    "make_sphinx_doc",
    "feature_flags_from_environ",
    "parse_feature_flags_string",
]



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
)


class FeatureFlag:
    """A descriptor object to mark attribute as a feature flag.

    Args:
        desciption: Human-readable description of the feature flag.
    """

    def __init__(self, description: str = "") -> None:
        self.desciption = description

    def __set_name__(self, owner: Any, name: str) -> None:
        self.name = name

    def __get__(self, obj: Any, cls: Optional[type] = None) -> bool:
        return getattr(obj, f"_{self.name}", False)

    def __set__(self, obj: Any, value: bool) -> None:
        setattr(obj, f"_{self.name}", value)


FeatureFlags_T = TypeVar("FeatureFlags_T", bound="FeatureFlags")


class FeatureFlagsMeta(type):
    """Metaclass for the feature flags.

    Args:
        make_doc: If provided, used to generate docstring for the resulting class.
    """

    __feature_flags__: Tuple[str, ...]

    def __new__(
        cls: Type[type],
        name: str,
        bases: Tuple[type, ...],
        namespace: Dict[str, Any],
        make_doc: Optional[Callable[[Mapping[str, FeatureFlag]], str]] = None,
    ) -> type:
        feature_flags: Dict[str, FeatureFlag] = {}
        annotations: Dict[str, type] = namespace.pop("__annotations__", {})

        for ns_name, ns_attr in list(namespace.items()):
            if isinstance(ns_attr, FeatureFlag):
                feature_flags[ns_name] = ns_attr
                annotations[ns_name] = bool

                del namespace[ns_name]

        if not ("__doc__" in namespace or make_doc is None):
            namespace["__doc__"] = make_doc(feature_flags)

        namespace["__feature_flags__"] = tuple(feature_flags)
        namespace["__slots__"] = tuple(feature_flags)  # TODO: merge slots
        namespace["__annotations__"] = annotations

        # https://github.com/python/mypy/issues/9282
        return super().__new__(cls, name, bases, namespace)  # type: ignore

    def _flags(cls: type) -> Iterable[str]:
        """Iterate through all defined feature flag names."""

        for xcls in cls.mro():
            if isinstance(xcls, FeatureFlagsMeta):
                yield from xcls.__feature_flags__  # TODO: avoid duplicates


class FeatureFlags(metaclass=FeatureFlagsMeta):
    def __init__(
        self,
        values: Optional[Mapping[str, bool]] = None,
        default: bool = False,
        default_key: str = "",
    ) -> None:
        """Initialize feature flags.

        Unknown feature flags are ignored.

        Args:
            values: Feature flag states.
            default: Default state for unset feature flags.
            default_key: Key from ``values`` with default value for unset flags.
        """

        names = set(self.__class__._flags())

        if values:
            if default_key:
                default = values.get(default_key, default)

            for name in names:
                setattr(self, name, values.get(name, default))
        else:
            for name in names:
                setattr(self, name, default)

    def _dict(self) -> Dict[str, bool]:
        """Get copy of the feature flags in a form of dictionary."""

        return {name: getattr(self, name) for name in self.__class__._flags()}

    def _copy(
        self: FeatureFlags_T, overrides: Optional[Mapping[str, bool]] = None
    ) -> FeatureFlags_T:
        """Get copy of the feature flags object.

        Args:
            overrides: Feature flag states to override.
        """

        values = self._dict()

        if overrides:
            values.update(overrides)

        return self.__class__(values)

    __copy__ = _copy

    def __or__(self: FeatureFlags_T, other: FeatureFlags_T) -> FeatureFlags_T:
        """Merge feature flags."""

        return self.__class__(
            {
                name: getattr(self, name) or getattr(other, name)
                for name in self.__class__._flags()
            }
        )

    def __ior__(self: FeatureFlags_T, other: FeatureFlags_T) -> FeatureFlags_T:
        """Merge feature flags in-place."""

        for name in self.__class__._flags():
            if getattr(other, name):
                setattr(self, name, True)

        return self

    def __repr__(self) -> str:
        """Return string representation of the feature flags object."""

        cls = self.__class__
        params = ", ".join(
            f"{name}={repr(getattr(self, name))}" for name in sorted(set(cls._flags()))
        )
        return f"{cls.__name__}({params})"



from typing import Any, Mapping, Tuple

from .core import FeatureFlag

TITLE = "Feature Flags class."
DESCRIPTION = "Unknown feature flags are ignored."
ALL_FLAG_DESCRIPTION = "Enable all feature flags by default."


def by_item_name(pair: Tuple[str, Any]) -> str:
    """Get first element of the ``pair``."""
    return pair[0]


def make_napoleon_doc(feature_flags: Mapping[str, FeatureFlag]) -> str:
    """Generate Google-style docstring for the feature flags class."""

    out = [
        TITLE,
        "",
        DESCRIPTION,
        "",
        "Args:",
        f"    all: {ALL_FLAG_DESCRIPTION}",
    ]

    for name, feature_flag in sorted(feature_flags.items(), key=by_item_name):
        out.append(f"    {name}: {feature_flag.desciption}")

    out.append("")

    return "\n".join(out)


def make_sphinx_doc(feature_flags: Mapping[str, FeatureFlag]) -> str:
    """Generate Sphinx-style docstring for the feature flags class."""

    out = [
        TITLE,
        "",
        DESCRIPTION,
        "",
        f":param all: {ALL_FLAG_DESCRIPTION}",
    ]

    for name, feature_flag in sorted(feature_flags.items(), key=by_item_name):
        out.append(f":param {name}: {feature_flag.desciption}")

    out.append("")

    return "\n".join(out)



from contextlib import ContextDecorator
from contextvars import ContextVar, Token
from typing import Any, Generic, List

from .core import FeatureFlags_T


class FeatureFlagsContext(ContextDecorator, Generic[FeatureFlags_T]):
    """A context manager wrapper around ``ContextVar[FeatureFlags]``.

    Args:
        var: Context variable with default flags set.
    """

    __slots__ = ("tokens", "var")
    tokens: List[Token[FeatureFlags_T]]
    var: ContextVar[FeatureFlags_T]

    def __init__(self, var: ContextVar[FeatureFlags_T]) -> None:
        self.tokens = []
        self.var = var

    def __enter__(self) -> FeatureFlags_T:
        """Set copy of the feature flags as the new context variable."""

        feature_flags = self.current._copy()

        self.tokens.append(self.var.set(feature_flags))

        return feature_flags

    def __exit__(self, *exc: Any) -> None:
        """Restore previous value of the context variable."""

        self.var.reset(self.tokens.pop())

    def get_current(self) -> FeatureFlags_T:
        """Get feature flags from the current context."""

        return self.var.get()

    current = property(get_current)



from os import environ
from typing import Iterable, Mapping, Optional, Tuple


def parse_feature_flags_string(
    s: Optional[str], /, sep: Optional[str] = None, neg: str = "!"
) -> Mapping[str, bool]:
    """Parse feature flags string into mapping name -> value.

    >>> parse_feature_flag_string("a !b")
    {'a': True, 'b': False}
    """
    if not s:
        return {}

    return {flag.lstrip(neg): not flag.startswith(neg) for flag in s.split(sep)}


def parse_bool(s: str) -> Optional[bool]:
    """Parse string as boolean.

    Values ``1, ``true``, ``t``, ``yes``, ``y``, ``on`` are parsed as ``True``.
    Values ``0``, ``false``, ``f``, ``no``, ``n``, ``off`` and empty string
    are parsed as ``False``. If none maches, return ``None``. Matching is
    case-insensitive, whitespaces at the beginning and at the end are stripped
    before performing the match.
    """

    s = s.strip().lower()

    if s in {"1", "true", "t", "yes", "y", "on"}:
        return True
    if s in {"0", "false", "f", "no", "n", "off", ""}:
        return False
    return None


def _feature_flags_from_environ(
    prefix: str,
    environ: Mapping[str, str],
) -> Iterable[Tuple[str, bool]]:
    """See :func:`feature_flags_from_environ`."""

    prefix_len = len(prefix)
    for key, value in environ.items():
        if key.startswith(prefix):
            parsed_value = parse_bool(value)
            if parsed_value is not None:
                yield key[prefix_len:].lower(), parsed_value


def feature_flags_from_environ(
    prefix: str,
    environ: Mapping[str, str] = environ,
) -> Mapping[str, bool]:
    """Extract feature flags mapping (name -> value) from env variables.

    Iterate through all environment variables matching given ``prefix``,
    try to parse them as bool. Resulting mapping keys are stripped from the
    ``prefix`` and transformed to lowercase. Values that does not look like a
    bool are ignored.

    Note:
        :func:`parse_bool` is used for parsing values.

    Args:
        prefix: Case sensitive Env variable prefix (e.g. ``MYPROJ_FEATURE_``).
        environ: Mapping with environment variables (defaults to ``os.environ``).
    """

    return dict(_feature_flags_from_environ(prefix, environ))
